from typing import Protocol
from enum import Enum


class Actionable(Protocol):
    """
    State or Rule.
    """

    number: int


class ActionType(Enum):
    SHIFT = "SHIFT"
    REDUCE = "REDUCE"
    ACCEPT = "ACCEPT"
    REJECT = "REJECT"

    def abbreviated(self) -> str:
        """
        Returns an abbreviated version of the action type.
        """
        if self == self.SHIFT:
            return "s"
        if self == self.REDUCE:
            return "r"
        if self == self.ACCEPT:
            return "a"
        if self == self.REJECT:
            return "e"


class Action:
    """
    Parsing action. Contains data relevant to the actionable object (State or Rule)
    and the relevant action to be applied.

    The `actionable_number` refers to the state number for `State` actionables and
    to the rule number for `Rule` actionables.
    """

    def __init__(self, actionable: Actionable, action_type: ActionType) -> None:
        self.actionable = actionable
        self.action_type = action_type

    @staticmethod
    def shift(state: Actionable) -> "Action":
        """
        Create a shift action to the given state.
        """
        return Action(state, ActionType.SHIFT)

    @staticmethod
    def reduce(rule: Actionable) -> "Action":
        """
        Create a reduce action of the given rule.
        """
        return Action(rule, ActionType.REDUCE)

    @staticmethod
    def accept() -> "Action":
        """
        Create an accept action.
        """
        return Action(None, ActionType.ACCEPT)

    def __repr__(self) -> str:
        return f"<Action: {self}>"

    def __str__(self) -> str:
        if self.action_type in {ActionType.ACCEPT, ActionType.REJECT}:
            return str(self.action_type.abbreviated())

        return f"{self.action_type.abbreviated()}{self.actionable.number}"

    def __hash__(self) -> int:
        return hash((self.actionable, self.action_type))

    def __eq__(self, other) -> bool:
        if not isinstance(other, Action):
            return False

        return (
            self.actionable == other.actionable
            and self.action_type == other.action_type
        )



from typing import Iterable

from syntactes import Rule, Token


class Grammar:
    """
    A grammar is a set of rules that describe a language.

    The only valid 'words' of the language are the given tokens.
    """

    def __init__(
        self, starting_rule: Rule, rules: Iterable[Rule], tokens: set[Token]
    ) -> None:
        """
        `starting_rule` should also be included in `rules`.
        """
        self.starting_rule = starting_rule
        self.rules = rules
        self.tokens = tokens



from typing import Optional

from syntactes.rule import Rule
from syntactes.token import Token


class LR0Item:
    """
    Item of LR0 parser. Contains rule and current position in rule.
    Current position is denoted with the dot '.'.
    """

    def __init__(self, rule: Rule, position: int) -> None:
        self.rule = rule
        self.position = position

    def dot_is_last(self) -> bool:
        """
        Returns True if the dot in the item is in the last position of the rhs,
        False otherwise.
        """
        return self.position == self.rule.rhs_len

    @property
    def after_dot(self) -> Optional[Token]:
        """
        Returns the symbol after the dot in the current item.
        If the dot is in the last position, returns None
        """
        if self.dot_is_last():
            return None

        return self.rule.rhs[self.position]

    @property
    def before_dot(self) -> Optional[Token]:
        """
        Returns the symbol before the dot in the current item.
        If the dot is in the first position, returns None.
        """
        if self.position == 0:
            return None

        return self.rule.rhs[self.position - 1]

    def __repr__(self) -> str:
        return f"<LR0Item: {self}>"

    def __str__(self) -> str:
        rhs = [s for s in self.rule.rhs]
        rhs.insert(self.position, ".")
        return f"{self.rule.lhs} -> " + " ".join(map(str, rhs))

    def __hash__(self) -> int:
        return hash((self.rule, self.position))

    def __eq__(self, other) -> bool:
        if not isinstance(other, LR0Item):
            return False

        return self.rule == other.rule and self.position == other.position



from typing import Iterable

from syntactes._item import LR0Item


class LR0State:
    """
    State of LR0 parser. A LR0 state is a set of LR0 items.
    """

    def __init__(self) -> None:
        self.number = None
        self.items = set()
        self.is_final = False

    @staticmethod
    def from_items(items: Iterable[LR0Item]) -> "LR0State":
        """
        Create an LR0 state from a set of LR0 items.
        """
        state = LR0State()
        {state.add_item(item) for item in items}

        return state

    def add_item(self, item: LR0Item) -> None:
        """
        Adds an item to the state.
        """
        self.items.add(item)

    def set_number(self, number: int) -> None:
        self.number = number

    def set_final(self) -> None:
        self.is_final = True

    def __repr__(self) -> str:
        return f"<LR0State: {self.number}>"

    def __str__(self) -> str:
        return f"{self.number}:" + "(" + ", ".join(map(str, self.items)) + ")"

    def __hash__(self) -> int:
        return hash(frozenset(self.items))

    def __eq__(self, other) -> bool:
        if not isinstance(other, LR0State):
            return False

        return self.items == other.items



from syntactes.token import Token


class Rule:
    """
    Production rule. Describes the break-down of a non-terminal symbol to
    other symbols.

    LHS -> RHS1 RHS2...
    """

    def __init__(self, number: int, lhs: Token, *args: tuple[Token]) -> None:
        self.number = number
        self.lhs = lhs
        self.rhs = args
        self.rhs_len = len(args)

    def __repr__(self) -> str:
        return f"<Rule: {self}>"

    def __str__(self) -> str:
        return f"{self.lhs} -> " + " ".join(map(str, self.rhs))

    def __hash__(self) -> int:
        return hash((self.lhs, self.rhs))

    def __eq__(self, other) -> bool:
        if not isinstance(other, Rule):
            return False

        return self.lhs == other.lhs and self.rhs == other.rhs
